
/* ====================================================================
 * Copyright (c) 1995-1998 The Apache Group.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer. 
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. All advertising materials mentioning features or use of this
 *    software must display the following acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * 4. The names "Apache Server" and "Apache Group" must not be used to
 *    endorse or promote products derived from this software without
 *    prior written permission. For written permission, please contact
 *    apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Group.
 *
 * 6. Redistributions of any form whatsoever must retain the following
 *    acknowledgment:
 *    "This product includes software developed by the Apache Group
 *    for use in the Apache HTTP server project (http://www.apache.org/)."
 *
 * THIS SOFTWARE IS PROVIDED BY THE APACHE GROUP ``AS IS'' AND ANY
 * EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE APACHE GROUP OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT,
 * STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
 * OF THE POSSIBILITY OF SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Group and was originally based
 * on public domain software written at the National Center for
 * Supercomputing Applications, University of Illinois, Urbana-Champaign.
 * For more information on the Apache Group and the Apache HTTP server
 * project, please see <http://www.apache.org/>.
 *
 */

#include "httpd.h"
#include "http_config.h"
#include "http_core.h"
#include "http_log.h"
#include "http_main.h"
#include "http_protocol.h"
#include "util_script.h"

#include <stdio.h>

#include <lber.h>
#include <ldap.h>

#define MOD_AUTH_LDAP_VERSION_S "2.4.2"

/*
** uncomment the following line if you're having problem. The debug messages
** will be written to the server error log
*/
/*
#define DEBUG_LDAP 1
*/

module MODULE_VAR_EXPORT ldap_auth_module;

/*
 * $Revision: 5 $
 * $Date: 6/08/03 12:55a $
 * $Author: Muquit $
 */

/*
**  mod_auth_ldap.c
**  Apache modue to use LDAP server for HTTP basic authentication.
**
**  Limitations and Comments:
**  It is simple and clean. works with Umich LDAP SDK and Netscape LDAP SDK
**
**  Development History:
**      who                  when           why
**   ma_muquit@fccc.edu      Sep-05-1998    first cut
**
**   ma_muquit@fccc.edu      Aug-19-1999    adding AuthLDAPAuthoritative
**        config file option, so that if the uid is not in the LDAP server,
**        control will be passed to lower modules.
**
**  Take.Vos@cable.a2000.nl  Sep-20-1999    fixed multiple groups free() too 
**       early. added partials DN to require group, instead of fixed group list
**
**  muquit@muquit.com        Feb-10-2000    replaced ap_getword() and my
**       rmallws() calls with  ap_getword_white(). Replaced ap_getword() and 
**       rmallws() calls for attributes with  ap_getword_conf() as 
**       suggested by Mark OLear <mgolear@ilstu.edu>. The reason to use
**       ap_getword_conf() was: if say require ou "Foo bar" is used, 
**       ap_geword() was tokenizing the string to Foo bar, but we want to
**       consider "Foo bar" as one string.
**
**  david@giffin.org         Jan-26-2001    Added a blind bind to see
**      if the user can bind with a dn of 'uid_attr=uid, base_dn' if
**      not we just keep going with the old scheme. This allows for
**      people to authenicate through a non-anonymous directory server
**      if it's set up consistantly. Also added a little better error when 
**      no user is found during the dn search.
**
**  muquit@muquit.com       Feb-03-2001     Added code to put LDAP_USER in the
**      environment if the user could authenticate successfully. Requested
**      by Ben Brewer <ben@polygon.net>.
**
**  muquit@muquit.com       Feb-05-2001     Possible memory leak. result in
**      ldap_search_s() was not freed by calling ldap_msgfree()
**
**  muquit@muquit.com       Mar-15-2001     Added Bind_DN, Bind_Pass. Added
**      env var MOD_AUTH_LDAP_VERSION. Release 2.3.
**
**  muquit@muquit.com       Mar-18-2001     added the filter option.
**
**  muquit@muquit.com       Apr-14-2001     The uids can have spaces now.
**      example: require user foo "john doe" bar 
**
*/


#ifndef LDAP_PORT
#define LDAP_PORT   389
#endif /* LDAP_PORT */

typedef struct _ldap_auth_config_rec
{
    LDAP
        *ld;

    char
        *ldap_server,
        *base_dn,
        *uid_attr;

    char
        *user_dn;

    char
        *bind_dn,
        *bind_pass;

    int
        ldap_port;

    int
        auth_ldapauthoritative;

    char
        *dummy;
} ldap_auth_config_rec;


static void *create_ldap_auth_dir_config(pool *p,char *d)
{
    ldap_auth_config_rec
        *cr;

    cr=(ldap_auth_config_rec *) ap_pcalloc(p,sizeof(ldap_auth_config_rec));
    if (cr == (ldap_auth_config_rec *) NULL)
        return (NULL);

    cr->ld=(LDAP *) NULL;

    cr->ldap_server=NULL; /* used to be localhost, mm, Aug-20-1999 */
    cr->base_dn=NULL;
    cr->user_dn=NULL;
    cr->bind_dn=NULL;
    cr->bind_pass=NULL;
    cr->uid_attr=ap_pstrdup(p,"uid");
    cr->ldap_port=LDAP_PORT;
    cr->auth_ldapauthoritative=1; /* fortress is secure by default */

    return (cr);
}

static const char *set_ldap_server(cmd_parms *cmd,ldap_auth_config_rec *cr,
                                   char *arg)
{

    cr->ldap_server=ap_pstrdup(cmd->pool,arg);

    return (NULL);
}

static const char *set_ldap_port(cmd_parms *cmd,ldap_auth_config_rec *cr,
                                 char *arg)
{
    cr->ldap_port=atoi(arg);

    return (NULL);
}

static const char *set_base_dn(cmd_parms *cmd,ldap_auth_config_rec *cr,
                                 char *arg)
{
    cr->base_dn=ap_pstrdup(cmd->pool,arg);
    
    return (NULL);
}

static const char *set_bind_dn(cmd_parms *cmd,ldap_auth_config_rec *cr,
                               char *arg)
{
    cr->bind_dn=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_bind_pass(cmd_parms *cmd,ldap_auth_config_rec *cr,
                               char *arg)
{
    cr->bind_pass=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_uid_attr(cmd_parms *cmd,ldap_auth_config_rec *cr,
                                char *arg)
{
    cr->uid_attr=ap_pstrdup(cmd->pool,arg);
    return (NULL);
}

static const char *set_ldapauthoritative(cmd_parms *cmd,ldap_auth_config_rec 
                                *cr,char *arg)
{
    char
        *a=ap_pstrdup(cmd->pool,arg);

    if (a != NULL)
    {
        if ((strcmp(a,"no") == 0) || 
            (strcmp(a,"off") == 0))
        {
#ifdef DEBUG_LDAP
           ap_log_error(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,NULL,
                "[mod_auth_ldap.c] -  AuthLDAPAuthoritative is off");
#endif /* DEBUG_LDAP */
            cr->auth_ldapauthoritative=0;
        }
        else 
        {
            if ((strcmp(a,"yes") == 0) || 
                (strcmp(a,"on") == 0))
            {
#ifdef DEBUG_LDAP
           ap_log_error(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,NULL,
                "[mod_auth_ldap.c] -  AuthLDAPAuthoritative is on");
#endif /* DEBUG_LDAP */
                cr->auth_ldapauthoritative=1;
#ifdef DEBUG_LDAP
            }
            else
            {
                ap_log_error(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,NULL,
                    "[mod_auth_ldap.c] -  AuthLDAPAuthoritative has unexpected argument");
#endif /* DEBUG_LDAP */
            }
        }
#ifdef DEBUG_LDAP
    }
    else
    {
        ap_log_error(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,NULL,
            "[mod_auth_ldap.c] -  AuthLDAPAuthoritative is missing its argument");
#endif /* DEBUG_LDAP */
    }
    return (NULL);
}

static const command_rec ldap_auth_cmds[]=
{
    {"LDAP_Server",set_ldap_server,NULL,OR_AUTHCFG,TAKE1,"LDAP server"},
    {"LDAP_Port",  set_ldap_port,  NULL,OR_AUTHCFG,TAKE1,"LDAP port"},
    {"Base_DN",    set_base_dn,    NULL,OR_AUTHCFG,TAKE1,"Base DN"},
    {"UID_Attr",   set_uid_attr,   NULL,OR_AUTHCFG,TAKE1,"uid"},
    {"Bind_DN",    set_bind_dn,    NULL,OR_AUTHCFG,TAKE1,"Bind DN"},
    {"Bind_Pass",  set_bind_pass,  NULL,OR_AUTHCFG,TAKE1,"Bind Password"},
    {"AuthLDAPAuthoritative",set_ldapauthoritative,NULL,OR_AUTHCFG,TAKE1,
        "Set to no to allow control to be passed along to lower modules if the uid is not known to this module"},
    {NULL}
};

/*
**  ldapFindUserDN()
**    return the DN of the user 
**
**  Parameters:
**  LDAP    *ld         valid handle to LDAP
**  char    *base_dn    LDAP base Distinguised Name
**  char    *uid_attrib LDAP uid attribute, e.g. "uid"
**  char    *userid     the userid to check
**  request_rect *r     needed for writing log in Win2K
**  char    *bind_dn    Bind_DN, can be NULL
**  char    *bind_pass  Bind_Pass, can be NULL
**
**  Return Values:
**      pointer to DN of the userid if found, NULL otherwise
**
**  Limitations and Comments:
**      ld must be valid.
**
**      any LDAP sdk should work. returns pointer to a malloc'd space, the 
**      caller is responsible to free it.
**
**  Development History:
**      who                  when           why
**      ma_muquit@fccc.edu   Nov-29-1998    first cut for my own web server
**                                          mhttpd
**      muquit@muquit.com    Mar-15-2001    bind with bind and pass if provided.
*/
static char *ldapFindUserDN(LDAP *ld,char *base_dn,char *uid_attrib,
        char *userid,request_rec *r,char *bind_dn,char *bind_pass)
{
    int
        count = 0;

    int
        bound=0,
        rc=(-1);

    LDAPMessage
        *result,
        *entry;

    char
        *dn,
        szfilter[256];

    result=(LDAPMessage *) NULL;
    entry=(LDAPMessage *) NULL;
    dn=(char *) NULL;

    /* prepare filter with UidAttr. */
    ap_snprintf(szfilter,sizeof(szfilter)-1,"(%s=%s)",uid_attrib,userid);

#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - filter: %s",szfilter);
#endif /* DEBUG_LDAP */

    /*
    ** if the bind DN and bind password is specified, try to bind with
    ** them first.
    */
    if (bind_dn != NULL && bind_pass != NULL)
    {
#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - trying to bind with bind DN \"%s\" and password (not shown)",bind_dn);
#endif /* DEBUG_LDAP */
        if (ldap_simple_bind_s(ld,bind_dn,bind_pass) == LDAP_SUCCESS)
        {
            bound=1;
#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - Bound successfully with DN \"%s\" and password (not shown)",bind_dn);

#endif /* DEBUG_LDAP */
        }
        else
        {
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - Warning: bind with DN \"%s\" and password (not shown) did not succeed",bind_dn);

        }
    }
    else
    {
#ifdef DEBUG_LDAP
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - Bind_DN/Bind_Pass is not provided");
#endif /* DEBUG_LDAP */
    }



    /* Bind anonymously now only if not bound yet. */
    if (bound == 0)
    {
        if (ldap_simple_bind_s(ld,base_dn,NULL) != LDAP_SUCCESS)
        {
#ifdef DEBUG_LDAP
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                    "[mod_auth_ldap.c] - ldap_simple_bind_s() failed");
#endif /* DEBUG_LDAP */

            return ((char *) NULL);
        }
    }

    /* search to find the dn for the user */
    rc=ldap_search_s(ld,
            base_dn,
            LDAP_SCOPE_SUBTREE,
            szfilter,
            NULL,
            0,
            &result);
    if (rc != LDAP_SUCCESS)
    {
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - ldap_search_s() failed");

        /* note: ldap_err2string() returns pointer to a static space */
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - Error: %s",ldap_err2string(rc));
        return (NULL);
    }

    /* found something, better be a single match */
    if ((count = ldap_count_entries(ld,result)) != 1)
    {
	    if (count > 0)
	    {
	        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - multiple matches found for: %s",userid);
	    }
	    else
	    {
            if (bound == 0)
            {
	            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                    "[mod_auth_ldap.c] - do we have anonymous access? no user found for: %s",userid);
            }
            else
            {
	            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                    "[mod_auth_ldap.c] - no user found for: %s",userid);
            }
	    }
        if (result != (LDAPMessage *) NULL)
            ldap_msgfree(result);

        return (NULL);
    }
    if ((entry=ldap_first_entry(ld,result)) == NULL)
    {

#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - %s=%s Unknown in LDAP server",uid_attrib,userid);
#endif /* DEBUG_LDAP */

        if (result != (LDAPMessage *) NULL)
            ldap_msgfree(result);
        return (NULL);
    }

    if ((dn=ldap_get_dn(ld,entry)) == NULL)
    {
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - could not obtain DN from LDAP for user %s",userid);

        return (NULL);
    }

#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - in  ldapFindUserDN() user DN: %s",dn);
#endif /* DEBUG_LDAP */
    return (dn); /* caller must free it */
}


/*
** remove all whitespace (leading or trailing) from a string.
** borrowed from public domain c snippets library.
** muquit, sep-07-1998
*/
static char *rmallws(char *str)
{
    char
        *obuf,
        *nbuf;

    if (str)
    {
        for (obuf=str, nbuf=str; *obuf; ++obuf)
        {
            if (!isspace(*obuf))
                *nbuf++ = *obuf;
        }
        *nbuf='\0';
    }
    return (str);
}

/* for my use only - muquit@muquit.com */
#ifdef MY_DEBUG
static void dumpResult(LDAP *ld,LDAPMessage *res,request_rec *r)
{
    LDAPMessage
        *ent;

    BerElement
        *ber;

    int
        i;

    char
        *attr,
        **vals;

ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
        "XXXXXXXX");
    for (ent=ldap_first_entry(ld,res); ent != NULL; ent=ldap_next_entry(ld,ent))
    {
        for (attr=ldap_first_attribute(ld,ent,&ber); attr != NULL;
                attr=ldap_next_attribute(ld,ent,ber))
        {
            if (strcasecmp(attr,"jpegphoto") != 0)
            {
                if ((vals=ldap_get_values(ld,ent,attr)) != NULL)
                {
                    for (i=0; vals[i] != NULL; i++)
                    {
ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
        "dump: %s=%s",attr,vals[i]);
                    }
                    if (vals)
                        ldap_value_free(vals);
                }
            }
        }
    }
}
#endif /* MY_DEBUG */

/*
** validate the user id via LDAP
** Returns:
**      DECLINED        if ldap server is not defined in config file
**      AUTH_REQUIRED   for any other kind of errors
**      OK              if authorized
*/
static int ldap_authenticate_basic_user(request_rec *r)
{
    ldap_auth_config_rec
        *cr;

    int
        rc,
        res;

    conn_rec
        *c;

     const char
        *sent_pw=NULL;

    char
        *dn,
        szfilter[256],
        sztdn[256];


    c=r->connection;
    cr=(ldap_auth_config_rec *)
        ap_get_module_config(r->per_dir_config,&ldap_auth_module);

    res=ap_get_basic_auth_pw(r,&sent_pw);
    if (res)
        return (res);

    /* if no user is specified, keep asking */
    if (c->user == NULL || *c->user == '\0')
    {
        if (cr->auth_ldapauthoritative)
        {
            ap_note_basic_auth_failure(r);
            return (AUTH_REQUIRED);
        }
        else
            return(DECLINED);

    }
    /* if no password is specified, keep asking */
    if ((sent_pw == NULL) || (*sent_pw == '\0'))
    {
        if (cr->auth_ldapauthoritative)
        {
            ap_note_basic_auth_failure(r);
            return (AUTH_REQUIRED);
        }
        else
            return(DECLINED);
    }

#ifdef DEBUG_LDAP
    ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
            "[mod_auth_ldap.c] - LDAP server=%s,Port=%d",cr->ldap_server,cr->ldap_port);
#endif /* DEBUG_LDAP */

    /* mm, aug-20-199 */
    if (cr->ldap_server == NULL)
        return (DECLINED);

    if (cr->ld != NULL)
    {
#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - closing connection to LDAP server: %s",
                cr->ldap_server);
#endif /* DEBUG_LDAP */
        ldap_unbind_s(cr->ld);
    }


#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "mod_auth_ldap.c] - opening connection to LDAP server: %s at port: %d",cr->ldap_server,cr->ldap_port);

#endif /* DEBUG_LDAP */

    /* connect to LDAP server */
    cr->ld=ldap_open(cr->ldap_server,cr->ldap_port);
    if (cr->ld == (LDAP *) NULL)
    {
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - Could not connect to LDAP server %s at port %d",cr->ldap_server,cr->ldap_port);

        ap_note_basic_auth_failure(r);
        return (AUTH_REQUIRED);
    }

    /* never follow referral */
    /** TODO: prove a config option */
    /*
    ldap_set_option(cr->ld,LDAP_OPT_REFERRALS,LDAP_OPT_OFF);
    */


/*
** we don't need the following anymore. If someone wants to bind with a DN
** and password they can do so now. Look above.
** muquit@muquit.com, Sun Mar 11 19:05:49 EST 2001
** 
*/

#if 0
    /*
     *    david@giffin.org   Jan-26-2001
     *    try to bind to <uid_attr>=<username>, <base_dn>
     */

    /*
    sprintf(tdn,"%s=%s, %s",cr->uid_attr,c->user,cr->base_dn);
    */
    /* don't use sprintf, muquit@muquit.com, Jan-27-2001 */
    ap_snprintf(sztdn,sizeof(sztdn),"%s=%s, %s",cr->uid_attr,c->user,cr->base_dn);

#ifdef DEBUG_LDAP
    ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
        "[mod_auth_ldap.c] - try to bind to %s ",sztdn);
#endif /* DEBUG_LDAP */

    if (ldap_simple_bind_s(cr->ld,sztdn,sent_pw) == LDAP_SUCCESS) 
    {

#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
            "[mod_auth_ldap.c] -   bind worked for: %s",c->user);
#endif /* DEBUG_LDAP */

        cr->dn=ap_pstrdup(r->pool,sztdn);
        return OK;
    }
#endif /* if 0 */


    /* now get the User DN */
    dn=ldapFindUserDN(cr->ld,cr->base_dn,cr->uid_attr,c->user,r,
            cr->bind_dn,cr->bind_pass);
    if (dn == (char *) NULL)
    {
#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - ldapFindUserDN() didn't return any DN for user \"%s\" with attr \"%s\"",c->user,cr->uid_attr);
#endif /* DEBUG_LDAP */

        /* pass control to lower modules if AuthLDAPAuthoritative=no */
        if (! (cr->auth_ldapauthoritative))
        {
#ifdef DEBUG_LDAP
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "returning control to lower modules");
#endif /* DEBUG_LDAP */
            return (DECLINED);
        }
        ap_note_basic_auth_failure(r);
        return (AUTH_REQUIRED);

    }

    /* store user DN, we'll use it later if needed*/
    cr->user_dn=ap_pstrdup(r->pool,dn);

#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
            "[mod_auth_ldap.c] - User DN: %s",dn);
#endif /* DEBUG_LDAP */

    /*
    ** now if we can successfully bind to LDAP server as the specified
    ** userid and passoword, the user is a valid one!
    */
    if (ldap_simple_bind_s(cr->ld,dn,sent_pw) != LDAP_SUCCESS)
    {
        ldap_unbind_s(cr->ld);
        cr->ld=NULL;
#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
            "[mod_auth_ldap.c] -  Invalid Login or Password for:%s",c->user);
#endif /* DEBUG_LDAP */

        ap_note_basic_auth_failure(r);
        return (AUTH_REQUIRED);
    }

    /* free things we won't need */
    if (dn != (char *) NULL)
    {
        /* 
         * if free is called, it dies in Windows 2000 if compiled with
         * iPlanet C SDK 5.08. ldap_memfree() is suggested use with most of 
         * the SDKs anyway.
         * - muquit@muquit.com Sep-24-2002 
         */
         
        /* (void) free(dn);*/
        ldap_memfree(dn);
        dn=NULL;
    }

    return (OK);
}

/* called to check to see if resource being requested requires authorization. */
static int ldap_check_user_access(request_rec *r)
{
    const array_header
        *reqs_arr;

    require_line
        *reqs;

    register int
        x;

    int
        rc,
        m;

    const char
        *t,
        *w;

    ldap_auth_config_rec
        *cr;

    conn_rec
        *c;
    
    char
        *attr,
        *user=r->connection->user;

    cr=(ldap_auth_config_rec *) NULL;
    reqs_arr=(array_header *)NULL;
    reqs=(require_line *) NULL;

    cr=(ldap_auth_config_rec *)
        ap_get_module_config(r->per_dir_config,&ldap_auth_module);
    reqs_arr=ap_requires(r);
    reqs=reqs_arr ? (require_line *) reqs_arr->elts : NULL;
    m=r->method_number;
    c=r->connection;

    /* mm, Aug-20-1999 */
    if (cr->ldap_server == NULL)
        return(DECLINED);

    if (cr->ld == (LDAP *) NULL)
    {
#ifdef DEBUG_LDAP
        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
             "[mod_auth_ldap.c] - no open LDAP connection");
#endif /* DEBUG_LDAP */

         return (DECLINED);
    }

    /* server cfg file must have atleast one require line */
    if (reqs_arr == NULL)
    {
            ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
             "[mod_auth_ldap.c] - no require line in config file");

        if (cr->ld != (LDAP *) NULL)
        {
            ldap_unbind(cr->ld);
            cr->ld=NULL;
        }
        return (DECLINED);
    }

    for (x=0; x < reqs_arr->nelts; x++)
    {
        if (!(reqs[x].method_mask & (1 << m)))
            continue;
        t=reqs[x].requirement;

        /*
        w=ap_getword(r->pool,&t,' ');
        rmallws((char *) w);
        */

        /*
        ** use ap_getword_white() instead of ap_getword() and my own rmallws
        ** muquit@muquit.com, Feb-10-2000
        */
        w=ap_getword_white(r->pool,&t);

        /*
        ** if "require valid-user" found in cfg, the user is authorized.
        ** we'r outta here with OK
        */
        if(strcmp(w,"valid-user") == 0)
        {

            if (cr->ld != (LDAP *) NULL)
            {
                ldap_unbind(cr->ld);
                cr->ld=NULL;
            }
            ap_table_setn(r->subprocess_env,"LDAP_USER",user);
            ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",
                    MOD_AUTH_LDAP_VERSION_S);

            /* remove later muquit@muquit.com */
#ifdef DEBUG_LDAP
           if (cr->user_dn != NULL)
           {
                ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                    "[mod_auth_ldap.c] - user dn obtained earlier: %s",
                    cr->user_dn);
           }
#endif /* DEBUG_LDAP */


            return (OK);
        }
        else if (strcmp(w,"user") == 0)
        {
            /*
            ** allow in only if the userid is one in the space separated list
            ** in the server cfg file. now t should have all the space 
            **  separated uids from cfg file 
            */
            while (*t != '\0')
            {
                /*
                ** go through all the uids and compare them with the uid
                ** from connection (which is in the LDAP server)
                */

                /*
                w=ap_getword(r->pool,&t,' ');
                rmallws((char *)w);
                */

                /*
                ** use ap_getword_white() instead of ap_getword() and my
                **  own rmallws, muquit@muquit.com, Feb-10-2000
                */
                /*
                w=ap_getword_white(r->pool,&t);
                */

                /*
                ** nope, use ap_getword_conf() instead, this way we can
                ** have uid with spaces in it, example:
                ** require user foo "j doe" bar
                ** muquit@muquit.com Sun Apr 15 10:19:06 EDT 2001 
                */
                w=ap_getword_conf(r->pool,&t);


#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - w=\"%s\"",w);
#endif /* DEBUG_LDAP */

                if (*w != '\0')
                {
                    if (strcmp(w,user) == 0)
                    {
#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - user %s is authorized",user);
#endif /* DEBUG_LDAP */

                        if (cr->ld != (LDAP *) NULL)
                        {
                            ldap_unbind(cr->ld);
                            cr->ld=NULL;
                        }
                        ap_table_setn(r->subprocess_env,"LDAP_USER",user);
                        ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",
                            MOD_AUTH_LDAP_VERSION_S);
                        return (OK);
                    }
                }
            }
            ap_note_basic_auth_failure(r);
            return (AUTH_REQUIRED);
        }
        else if (strcmp(w,"filter") == 0)
        {
            int
                rc;

            LDAPMessage
                *result;
            char 
                *filter=NULL;

            /*
            ** We'll check if the user match the attributes in filter.
            ** We're going to use the user_dn obtained earlier, and we'll
            ** use the same open connection.
            */
            if (cr->ld == (LDAP *) NULL)
            {
                ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                        "[mod_auth_ldap.c] - ldap_check_user_access() no open LDAP connection");
                ap_note_basic_auth_failure(r);
                return (AUTH_REQUIRED);
            }

            if (cr->user_dn == NULL)
            {
                ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                        "[mod_auth_ldap.c] - ldap_check_user_access() no use DN available");
                ap_note_basic_auth_failure(r);
                return (AUTH_REQUIRED);
            }

            filter=ap_getword_conf(r->pool,&t);
            rc=ldap_search_s(cr->ld,
                        cr->user_dn,
                        LDAP_SCOPE_SUBTREE,
                        filter,
                        NULL,
                        0,
                        &result);
            if (ldap_first_entry(cr->ld,result) == NULL)
            {
                if (result != NULL)
                    ldap_msgfree(result);

#ifdef DEBUG_LDAP
                ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                        "[mod_auth_ldap.c] - ldap_check_user_access() filter %s did not match  for user %s",filter,user);
#endif /* DEBUG_LDAP */

                ap_note_basic_auth_failure(r);
                return (AUTH_REQUIRED);
            }

            if (rc == LDAP_SUCCESS && result != NULL)
            {

#ifdef DEBUG_LDAP
                ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                        "[mod_auth_ldap.c] - ldap_check_user_access() filter %s matched for user %s",filter,user);
#endif /* DEBUG_LDAP */

#ifdef MY_DEBUG /* for me only */
            dumpResult(cr->ld,result,r);
#endif /* MY_DEBUG */

                if (result != NULL)
                    ldap_msgfree(result);

                if (cr->ld != (LDAP *) NULL)
                {
                    ldap_unbind(cr->ld);
                    cr->ld=NULL;
                }
                ap_table_setn(r->subprocess_env,"LDAP_USER",user);
                ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",
                    MOD_AUTH_LDAP_VERSION_S);
                return (OK);
            }

            if (result != NULL)
                ldap_msgfree(result);
            
#ifdef DEBUG_LDAP
                ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                        "[mod_auth_ldap.c] - ldap_check_user_access() filter %s did not match  for user %s",filter,user);
#endif /* DEBUG_LDAP */
            ap_note_basic_auth_failure(r);
            return (AUTH_REQUIRED);
        }
        else if (strcmp(w,"group") == 0)
        {
            char 
                *group_attrib=NULL,
                *group_name=NULL,
                szgdn[1024];

            group_attrib=ap_pstrdup(r->pool,w);
            rmallws(group_attrib);

#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - ldap attribute=%s",group_attrib);
#endif /* DEBUG_LDAP */


           /*
           ** XXX why not using the dn we found in user access??
           ** muquit@muquit.com 
           */
#ifdef DEBUG_LDAP
           if (cr->user_dn != NULL)
           {
                ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                    "[mod_auth_ldap.c] - using user DN obtained earlier: %s",
                    cr->user_dn);
           }
#endif /* DEBUG_LDAP */

            /* get DN */
           /*
            dn=ldapFindUserDN(cr->ld,cr->base_dn,cr->uid_attr,c->user,r,
                    cr->bind_dn,cr->bind_pass);
           */
            if (cr->user_dn == (char *) NULL)
            {
#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] -  no dn found for user in group attrib");
#endif /* DEBUG_LDAP */

                /* pass control to lower modules if AuthLDAPAuthoritative=no */
                if (! (cr->auth_ldapauthoritative))
                {
#ifdef DEBUG_LDAP
    ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
        "[mod_auth_ldap.c] - returning control to lower modules");
#endif /* DEBUG_LDAP */
                    return (DECLINED);
#ifdef DEBUG_LDAP
                }
                else
                {
                    ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                        "[mod_auth_ldap.c] - AuthLDAPAuthoritative=on");
#endif /* DEBUG_LDAP */
                }
                if (cr->ld != (LDAP *) NULL)
                {
                    ldap_unbind(cr->ld);
                    cr->ld=NULL;
                }
                ap_note_basic_auth_failure(r);
                return (AUTH_REQUIRED);
            }

            while (*t != '\0')
            {
                /*
                w=ap_getword(r->pool,&t,' ');
                rmallws((char *) w);
                */

                /*
                ** use ap_getword_conf() instead of the above. ap_getword()
                ** tokenizes the string, so a group like "foo group" can not
                ** be used. -- muquit@muquit.com, Feb-10-2000
                */
                w=ap_getword_conf(r->pool,&t);
                if (*w != '\0')
                {
                    ap_snprintf(szgdn,sizeof(szgdn)-1,
                        "%s,%s",w,cr->base_dn);
#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - gdn=%s",szgdn);
#endif /* DEBUG_LDAP */

                    rc=ldap_compare_s(cr->ld,
                            szgdn,
                            "uniquemember",
                            cr->user_dn);

                    if (rc == LDAP_COMPARE_TRUE)
                    {
#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - user %s belongs to group %s",
                            c->user,w);
#endif /* DEBUG_LDAP */
           /*
                        (void) free ((char *) dn);
                        dn=NULL;
            */
                        ldap_unbind(cr->ld);
                        cr->ld=NULL;
                        ap_table_setn(r->subprocess_env,"LDAP_USER",user);
                        ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",
                            MOD_AUTH_LDAP_VERSION_S);
                        return (OK);
                    }
                    else if (rc == LDAP_COMPARE_FALSE)
                    {
#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - user %s does not belong to group %s",
                 c->user,w);
#endif /* DEBUG_LDAP */
                    }
                    else
                    {
#ifdef DEBUG_LDAP
                       ldap_perror(cr->ld,"[mod_auth_ldap.c] Error:");
#endif /* DEBUG_LDAP */
                    }
                }
            }   /* while */

            /*
            if (dn != NULL)
            {
                (void) free ((char *) dn);
                dn=NULL;
            }
            */

            if (! (cr->auth_ldapauthoritative))
            {
#ifdef DEBUG_LDAP
                    ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                        "returning control to lower modules");
#endif /* DEBUG_LDAP */
                return (DECLINED);
            }
            ap_note_basic_auth_failure(r);
            return (AUTH_REQUIRED);
        }
        else
        {
            /*
            ** else allow in only if the ldap attribute is present in the LDAP
            ** server for the user
            */
            if (*w != '\0')
                attr=ap_pstrdup(r->pool,w);
#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - ldap attribute=%s",w);
#endif /* DEBUG_LDAP */


            /* go through the attribute values and check if the user has it */
            while (*t != '\0')
            {
                /*
                w=ap_getword(r->pool,&t,' ');
                rmallws((char *) w);
                */
                /*
                ** use ap_getword_conf() instead of the above. ap_getword()
                ** tokenizes string, so a attribute like "Foo org" can not
                ** be used.  -- muquit@muquit.com, Feb-10-2000
                */
                w=ap_getword_conf(r->pool,&t);
                if (*w != '\0')
                {
                    rc=ldap_compare_s(cr->ld,cr->user_dn,attr,w);
                    if (rc == LDAP_COMPARE_TRUE)
                    {
#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - user \"%s\" has attribute %s=\"%s\"",user,attr,w);
#endif /* DEBUG_LDAP */

                        ldap_unbind(cr->ld);
                        cr->ld=NULL;
                        ap_table_setn(r->subprocess_env,"LDAP_USER",user);
                        ap_table_setn(r->subprocess_env,"MOD_AUTH_LDAP_VERSION",
                            MOD_AUTH_LDAP_VERSION_S);
                        return (OK);
                    }
                    else
                    {
#ifdef DEBUG_LDAP
                        ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                        "[mod_auth_ldap.c] - user %s does not have attibute  %s=%s",user,attr,w);
#endif /* DEBUG_LDAP */
                    }
                }
            }
            ap_note_basic_auth_failure(r);
            return AUTH_REQUIRED;
        }
    }

    if (cr->ld != (LDAP *) NULL)
    {
#ifdef DEBUG_LDAP
           ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
                "[mod_auth_ldap.c] - unbinding LDAP");
#endif /* DEBUG_LDAP */

        ldap_unbind(cr->ld);
        cr->ld=NULL;
    }

#ifdef DEBUG_LDAP
    ap_log_rerror(APLOG_MARK,APLOG_NOERRNO | APLOG_ERR,r,
           "[mod_auth_ldap.c] - user \"%s\" is not allowd",user);
#endif /* DEBUG_LDAP */


    return (DECLINED);
}

/*--------------------------------------------------------------------------*/
/* Finally, the list of callback routines and data structures that          */
/* provide the hooks into our module from the other parts of the server.    */
/*--------------------------------------------------------------------------*/
module ldap_auth_module =
{
    STANDARD_MODULE_STUFF,
    NULL,                         /* module initializer */
    create_ldap_auth_dir_config,  /* per-directory config creator */
    NULL,                         /* dir config merger */
    NULL,                         /* server config creator */
    NULL,                         /* server config merger */
    ldap_auth_cmds,               /* command table */
    NULL,                         /* [7] list of handlers */
    NULL,                         /* [2] filename-to-URI translation */
    ldap_authenticate_basic_user, /* [5] check/validate user_id */
    ldap_check_user_access,       /* [6] check auth */
    NULL,                         /* [4] check access by host address */
    NULL,                         /* [7] MIME type checker/setter */
    NULL,                         /* [8] fixups */
    NULL,                         /* [10] logger */
    NULL,                         /* [3] header parser */
    NULL,                         /* process initializer */
    NULL,                         /* process exit/cleanup */
    NULL                          /* [1] post read_request handling */
};
